from .el_to_tex.adapter import MathMLElementToLatexConverterAdapter
from .xml_to_mathml.factory import make_xml_to_mathml_adapter


class MathMLToLaTeX:
    @staticmethod
    def convert(mathml: str) -> str:
        mathml_elements = make_xml_to_mathml_adapter().convert(mathml)
        mathml_elements_to_latex_converters = [
            MathMLElementToLatexConverterAdapter().to_latex_converter(mathml_element)
            for mathml_element in mathml_elements
        ]
        return "".join(
            [
                to_latex_converters.convert()
                for to_latex_converters in mathml_elements_to_latex_converters
            ]
        ).strip()



from typing import Optional, List
import re
from enum import Enum

from ..syntax import (
    HashUTF8ToLtXConverter,
    all_math_symbols_by_char,
    all_math_symbols_by_glyph,
    math_number_by_glyph,
    all_math_operators_by_char,
    all_math_operators_by_glyph,
)
from .protocols import (
    ToLaTeXConverter,
    MathMLElement,
    InvalidNumberOfChildrenError,
    MIMathMlElement,
)
from .services.normalizer import WhiteSpaceNormalizer
from .services.wrapper import (
    ParenthesisWrapper,
    BracketWrapper,
    GenericWrapper,
    JoinWithManySeparators,
)


class Math(ToLaTeXConverter):
    def __init__(self, math_element: MathMLElement, adapter):
        self._white_space_normalizer = WhiteSpaceNormalizer()
        self._adapter = adapter
        self._math_element = math_element

    def convert(self) -> str:
        raw_tex = " ".join(
            [
                self._adapter.to_latex_converter(child).convert()
                for child in self._math_element.children
            ]
        )
        return self._white_space_normalizer.normalize(raw_tex)


class Void(ToLaTeXConverter):
    def __init__(self, math_element: MathMLElement):
        self._math_element = math_element

    def convert(self) -> str:
        return ""


class MI(ToLaTeXConverter):
    def __init__(self, math_element: MathMLElement):
        self._utf8_converter = HashUTF8ToLtXConverter()
        self._normalizer = WhiteSpaceNormalizer()
        self._math_element = math_element

    def convert(self) -> str:
        normalized_value = self._normalizer.normalize(self._math_element.value)
        if normalized_value == " ":
            return Character.apply(normalized_value)

        trimmed_value = normalized_value.strip()
        converted_char = Character.apply(trimmed_value)

        parsed_char = self._utf8_converter.convert(converted_char)
        if parsed_char != converted_char:
            return parsed_char

        return self._wrap_in_math_variant(
            converted_char, self._get_math_variant(self._math_element.attributes)
        )

    def _get_math_variant(self, attributes: dict) -> str:
        if not attributes or "mathvariant" not in attributes:
            return ""
        return attributes["mathvariant"]

    def _wrap_in_math_variant(self, value: str, math_variant: str) -> str:
        if math_variant == "bold":
            return f"\\mathbf{{{value}}}"
        if math_variant == "italic":
            return f"\\mathit{{{value}}}"
        if math_variant == "bold-italic":
            return f"\\mathbf{{\\mathit{{{value}}}}}"
        if math_variant == "double-struck":
            return f"\\mathbb{{{value}}}"
        if math_variant == "bold-fraktur":
            return f"\\mathbf{{\\mathfrak{{{value}}}}}"
        if math_variant == "script":
            return f"\\mathcal{{{value}}}"
        if math_variant == "bold-script":
            return f"\\mathbf{{\\mathcal{{{value}}}}}"
        if math_variant == "fraktur":
            return f"\\mathfrak{{{value}}}"
        if math_variant == "sans-serif":
            return f"\\mathsf{{{value}}}"
        if math_variant == "bold-sans-serif":
            return f"\\mathbf{{\\mathsf{{{value}}}}}"
        if math_variant == "sans-serif-italic":
            return f"\\mathsf{{\\mathit{{{value}}}}}"
        if math_variant == "sans-serif-bold-italic":
            return f"\\mathbf{{\\mathsf{{\\mathit{{{value}}}}}}}"
        if math_variant == "monospace":
            return f"\\mathtt{{{value}}}"
        return value


class Character:
    def __init__(self, value: str):
        self.value = value

    @staticmethod
    def apply(value: str) -> str:
        return Character(value)._apply()

    def _apply(self) -> str:
        return (
            self._find_by_character()
            or self._find_by_glyph()
            or self._find_by_number()
            or HashUTF8ToLtXConverter().convert(self.value)
        )

    def _find_by_character(self) -> str:
        return all_math_symbols_by_char.get(self.value, "")

    def _find_by_glyph(self) -> str:
        return all_math_symbols_by_glyph.get(self.value, "")

    def _find_by_number(self) -> str:
        return math_number_by_glyph.get(self.value, "")


class MO(ToLaTeXConverter):
    def __init__(self, math_element: MathMLElement):
        self.normalizer = WhiteSpaceNormalizer()
        self._math_element = math_element

    def convert(self) -> str:
        normalized_value = self.normalizer.normalize(self._math_element.value)
        trimmed_value = normalized_value.strip()

        return Operator.operate(trimmed_value)


class Operator:
    def __init__(self, value: str):
        self._value = value

    @staticmethod
    def operate(value: str) -> str:
        return Operator(value)._operate()

    def _operate(self) -> str:
        # Attempt to find the LaTeX command by character
        latex_command = self._find_by_character()
        if latex_command:
            return latex_command

        # Attempt to find the LaTeX command by glyph
        latex_command = self._find_by_glyph()
        if latex_command:
            return latex_command

        # Attempt to find the LaTeX command by number
        latex_command = self._find_by_number()
        if latex_command:
            return latex_command

        # Fallback to converting using HashUTF8ToLtXConverter
        return HashUTF8ToLtXConverter().convert(self._value)

    def _find_by_character(self) -> Optional[str]:
        return all_math_operators_by_char.get(self._value)

    def _find_by_glyph(self) -> Optional[str]:
        return all_math_operators_by_glyph.get(self._value)

    def _find_by_number(self) -> Optional[str]:
        return math_number_by_glyph.get(self._value)


class MN(ToLaTeXConverter):
    def __init__(self, math_element: MathMLElement):
        self._normalizer = WhiteSpaceNormalizer()
        self._math_element = math_element

    def convert(self) -> str:
        normalized_value = self._normalizer.normalize(self._math_element.value).strip()
        converted_value = math_number_by_glyph.get(normalized_value, "")

        return converted_value or normalized_value


class MSup(ToLaTeXConverter):
    def __init__(self, math_element: MathMLElement, adapter):
        self._math_element = math_element
        self._adapter = adapter
        self._parenthesis_wrapper = ParenthesisWrapper()
        self._bracket_wrapper = BracketWrapper()

    def convert(self) -> str:
        name = self._math_element.name
        children = self._math_element.children
        children_length = len(children)

        if children_length != 2:
            raise InvalidNumberOfChildrenError(name, 2, children_length)

        base_child = children[0]
        exponent_child = children[1]

        return f"{self._handle_base_child(base_child)}^{self._handle_exponent_child(exponent_child)}"

    def _handle_base_child(self, base: MathMLElement) -> str:
        base_children = base.children
        base_str = self._adapter.to_latex_converter(base).convert()

        if len(base_children) <= 1:
            return base_str

        return self._parenthesis_wrapper.wrap_if_more_than_one_char(base_str)

    def _handle_exponent_child(self, exponent: MathMLElement) -> str:
        exponent_str = self._adapter.to_latex_converter(exponent).convert()

        return self._bracket_wrapper.wrap(exponent_str)


class GenericSpacingWrapper(ToLaTeXConverter):
    def __init__(self, math_element: MathMLElement, adapter):
        self._math_element = math_element
        self._adapter = adapter

    def convert(self) -> str:
        return " ".join(
            [
                self._adapter.to_latex_converter(child).convert()
                for child in self._math_element.children
            ]
        )


class MSqrt(ToLaTeXConverter):
    def __init__(self, math_element: MathMLElement, adapter):
        self._math_element = math_element
        self._adapter = adapter

    def convert(self) -> str:
        latex_joined_children = " ".join(
            [
                self._adapter.to_latex_converter(child).convert()
                for child in self._math_element.children
            ]
        )

        return f"\\sqrt{{{latex_joined_children}}}"


class Separators:
    def __init__(self, op: str, close: str):
        self._open = op
        self._close = close

    def wrap(self, string: str) -> str:
        return GenericWrapper(self._open, self._close).wrap(string)

    def are_parentheses(self) -> bool:
        return self._compare("(", ")")

    def are_square_brackets(self) -> bool:
        return self._compare("[", "]")

    def are_brackets(self) -> bool:
        return self._compare("{", "}")

    def are_divides(self) -> bool:
        return self._compare("|", "|")

    def are_parallels(self) -> bool:
        return self._compare("||", "||")

    def are_not_equal(self) -> bool:
        return self._open != self._close

    def _compare(self, open_to_compare: str, close_to_compare: str) -> bool:
        return self._open == open_to_compare and self._close == close_to_compare


class Vector:
    def __init__(self, op: str, close: str, separators: List[str]):
        self._open = op if op != "" else "("
        self._close = close if close != "" else ")"
        self._separators = separators

    def apply(self, latex_contents):
        content_without_wrapper = JoinWithManySeparators(self._separators).join(
            latex_contents
        )
        return GenericWrapper(self._open, self._close).wrap(content_without_wrapper)


class Matrix:
    def __init__(self, op: str, close: str):
        self._separators = Separators(op, close)
        self._generic_command = "matrix"

    def apply(self, latex_contents):
        command = self._command
        matrix = f'\\begin{{{command}}}\n{"".join(latex_contents)}\n\\end{{{command}}}'

        return (
            self._separators.wrap(matrix)
            if command == self._generic_command
            else matrix
        )

    @property
    def _command(self):
        if self._separators.are_parentheses():
            return "pmatrix"
        if self._separators.are_square_brackets():
            return "bmatrix"
        if self._separators.are_brackets():
            return "Bmatrix"
        if self._separators.are_divides():
            return "vmatrix"
        if self._separators.are_parallels():
            return "Vmatrix"
        if self._separators.are_not_equal():
            return self._generic_command
        return "bmatrix"


class MFenced(ToLaTeXConverter):
    def __init__(self, math_element: MathMLElement, adapter):
        self._math_element = math_element
        self._adapter = adapter
        self._open = self._math_element.attributes.get("open", "")
        self._close = self._math_element.attributes.get("close", "")
        self._separators = list(self._math_element.attributes.get("separators", ""))

    def convert(self) -> str:
        latex_children = [
            self._adapter.to_latex_converter(child).convert()
            for child in self._math_element.children
        ]

        if self._is_there_relative_of_name(self._math_element.children, "mtable"):
            return Matrix(self._open, self._close).apply(latex_children)

        return Vector(self._open, self._close, self._separators).apply(latex_children)

    def _is_there_relative_of_name(self, mathml_elements, element_name):
        return any(
            [
                child.name == element_name
                or self._is_there_relative_of_name(child.children, element_name)
                for child in mathml_elements
            ]
        )


class MTable(ToLaTeXConverter):
    def __init__(self, math_element: MathMLElement, adapter):
        self._math_element = math_element
        self._adapter = adapter
        self._add_flag_recursive_if_name(
            self._math_element.children, "mtable", "innerTable"
        )

    def convert(self) -> str:
        table_content = " \\\\\n".join(
            [
                self._adapter.to_latex_converter(child).convert()
                for child in self._math_element.children
            ]
        )

        return (
            self._wrap(table_content) if self._has_flag("innerTable") else table_content
        )

    def _wrap(self, latex: str) -> str:
        return f"\\begin{{matrix}}{latex}\\end{{matrix}}"

    def _add_flag_recursive_if_name(self, mathml_elements, name, flag):
        for mathml_element in mathml_elements:
            if mathml_element.name == name:
                mathml_element.attributes[flag] = flag
            self._add_flag_recursive_if_name(mathml_element.children, name, flag)

    def _has_flag(self, flag: str) -> bool:
        return bool(self._math_element.attributes.get(flag))


class MTr(ToLaTeXConverter):
    def __init__(self, math_element: MathMLElement, adapter):
        self._math_element = math_element
        self._adapter = adapter

    def convert(self) -> str:
        return " & ".join(
            [
                self._adapter.to_latex_converter(child).convert()
                for child in self._math_element.children
            ]
        )


class MSub(ToLaTeXConverter):
    def __init__(self, math_element: MathMLElement, adapter):
        self._math_element = math_element
        self._adapter = adapter
        self._parenthesis_wrapper = ParenthesisWrapper()
        self._bracket_wrapper = BracketWrapper()

    def convert(self) -> str:
        name = self._math_element.name
        children = self._math_element.children
        children_length = len(children)

        if children_length != 2:
            raise InvalidNumberOfChildrenError(name, 2, children_length)

        base_child = children[0]
        subscript_child = children[1]

        return f"{self._handle_base_child(base_child)}_{self._handle_subscript_child(subscript_child)}"

    def _handle_base_child(self, base: MathMLElement) -> str:
        base_children = base.children
        base_str = self._adapter.to_latex_converter(base).convert()

        if len(base_children) <= 1:
            return base_str

        return self._parenthesis_wrapper.wrap_if_more_than_one_char(base_str)

    def _handle_subscript_child(self, subscript: MathMLElement) -> str:
        subscript_str = self._adapter.to_latex_converter(subscript).convert()

        return self._bracket_wrapper.wrap(subscript_str)


class MFrac(ToLaTeXConverter):
    def __init__(self, math_element: MathMLElement, adapter):
        self._math_element = math_element
        self._adapter = adapter
        self._parenthesis_wrapper = ParenthesisWrapper()

    def convert(self) -> str:
        children = self._math_element.children
        name = self._math_element.name
        children_length = len(children)

        if children_length != 2:
            raise InvalidNumberOfChildrenError(name, 2, children_length)

        num = self._adapter.to_latex_converter(children[0]).convert()
        den = self._adapter.to_latex_converter(children[1]).convert()

        if self._is_bevelled():
            return f"{self._wrap_if_more_than_one_char(num)}/{self._wrap_if_more_than_one_char(den)}"

        return f"\\frac{{{num}}}{{{den}}}"

    def _wrap_if_more_than_one_char(self, val: str) -> str:
        return self._parenthesis_wrapper.wrap_if_more_than_one_char(val)

    def _is_bevelled(self) -> bool:
        return bool(self._math_element.attributes.get("bevelled"))


class MRoot(ToLaTeXConverter):
    def __init__(self, math_element: MathMLElement, adapter):
        self._math_element = math_element
        self._adapter = adapter

    def convert(self) -> str:
        name = self._math_element.name
        children = self._math_element.children
        children_length = len(children)

        if children_length != 2:
            raise InvalidNumberOfChildrenError(name, 2, children_length)

        content = self._adapter.to_latex_converter(children[0]).convert()
        root_index = self._adapter.to_latex_converter(children[1]).convert()

        return f"\\sqrt[{root_index}]{{{content}}}"


class MAction(ToLaTeXConverter):
    def __init__(self, math_element: MathMLElement, adapter):
        self._math_element = math_element
        self._adapter = adapter

    def convert(self) -> str:
        children = self._math_element.children

        if self._is_toggle():
            return " \\Longrightarrow ".join(
                [
                    self._adapter.to_latex_converter(child).convert()
                    for child in children
                ]
            )

        return self._adapter.to_latex_converter(children[0]).convert()

    def _is_toggle(self) -> bool:
        action_type = self._math_element.attributes.get("actiontype")
        return action_type == "toggle" or not action_type


class MEnclose(ToLaTeXConverter):
    def __init__(self, math_element: MathMLElement, adapter):
        self._math_element = math_element
        self._adapter = adapter
        self._notation = self._math_element.attributes.get("notation", "longdiv")

    def convert(self) -> str:
        latex_joined_children = " ".join(
            [
                self._adapter.to_latex_converter(child).convert()
                for child in self._math_element.children
            ]
        )

        if self._notation == "actuarial":
            return f"\\overline{{\\left.{latex_joined_children}\\right|}}"
        if self._notation == "radical":
            return f"\\sqrt{{{latex_joined_children}}}"
        if self._notation in ["box", "roundedbox", "circle"]:
            return f"\\boxed{{{latex_joined_children}}}"
        if self._notation == "left":
            return f"\\left|{latex_joined_children}"
        if self._notation == "right":
            return f"{latex_joined_children}\\right|"
        if self._notation == "top":
            return f"\\overline{{{latex_joined_children}}}"
        if self._notation == "bottom":
            return f"\\underline{{{latex_joined_children}}}"
        if self._notation == "updiagonalstrike":
            return f"\\cancel{{{latex_joined_children}}}"
        if self._notation == "downdiagonalstrike":
            return f"\\bcancel{{{latex_joined_children}}}"
        if self._notation == "updiagonalarrow":
            return f"\\cancelto{{}}{{{latex_joined_children}}}"
        if self._notation in ["verticalstrike", "horizontalstrike"]:
            return f"\\hcancel{{{latex_joined_children}}}"
        if self._notation == "madruwb":
            return f"\\underline{{{latex_joined_children}\\right|}}"
        if self._notation == "phasorangle":
            return f"{{\\angle \\underline{{{latex_joined_children}}}}}"
        return f"\\overline{{\\left.\\right){latex_joined_children}}}"


class MError(ToLaTeXConverter):
    def __init__(self, math_element: MathMLElement, adapter):
        self._math_element = math_element
        self._adapter = adapter

    def convert(self) -> str:
        latex_joined_children = " ".join(
            [
                self._adapter.to_latex_converter(child).convert()
                for child in self._math_element.children
            ]
        )

        return f"\\color{{red}}{{{latex_joined_children}}}"


class MSubsup(ToLaTeXConverter):
    def __init__(self, math_element: MathMLElement, adapter):
        self._math_element = math_element
        self._adapter = adapter
        self._parenthesis_wrapper = ParenthesisWrapper()
        self._bracket_wrapper = BracketWrapper()

    def convert(self) -> str:
        name = self._math_element.name
        children = self._math_element.children
        children_length = len(children)

        if children_length != 3:
            raise InvalidNumberOfChildrenError(name, 3, children_length)

        base_child = children[0]
        subscript_child = children[1]
        superscript_child = children[2]

        return f"{self._handle_base_child(base_child)}_{self._handle_subscript_child(subscript_child)}^{self._handle_superscript_child(superscript_child)}"

    def _handle_base_child(self, base: MathMLElement) -> str:
        base_children = base.children
        base_str = self._adapter.to_latex_converter(base).convert()

        if len(base_children) <= 1:
            return base_str

        return self._parenthesis_wrapper.wrap_if_more_than_one_char(base_str)

    def _handle_subscript_child(self, subscript: MathMLElement) -> str:
        subscript_str = self._adapter.to_latex_converter(subscript).convert()

        return self._bracket_wrapper.wrap(subscript_str)

    def _handle_superscript_child(self, superscript: MathMLElement) -> str:
        superscript_str = self._adapter.to_latex_converter(superscript).convert()

        return self._bracket_wrapper.wrap(superscript_str)


class TextCommand:
    def __init__(self, math_variant: str):
        self._math_variant = math_variant or "normal"

    def apply(self, value: str) -> str:
        acc = value
        for command in self._commands:
            acc = f"{command}{{{acc}}}"

        return acc

    @property
    def _commands(self) -> List[str]:
        if self._math_variant == "bold":
            return ["\\textbf"]
        if self._math_variant == "italic":
            return ["\\textit"]
        if self._math_variant == "bold-italic":
            return ["\\textit", "\\textbf"]
        if self._math_variant == "double-struck":
            return ["\\mathbb"]
        if self._math_variant == "monospace":
            return ["\\mathtt"]
        if self._math_variant in ["bold-fraktur", "fraktur"]:
            return ["\\mathfrak"]
        return ["\\text"]


class Char:
    def __init__(self, value: str, is_alphanumeric: bool):
        self.value = value
        self.is_alphanumeric = is_alphanumeric


class MText(ToLaTeXConverter):
    def __init__(self, math_element: MathMLElement):
        self._math_element = math_element

    def convert(self) -> str:
        attributes = self._math_element.attributes
        value = self._math_element.value

        chars = [
            (
                Char(char, True)
                if re.match(r"^[a-zA-Z0-9]$", char) or char == " "
                else Char(char, False)
            )
            for char in value
        ]

        merged_chars = []
        for char in chars:
            if char.is_alphanumeric:
                last_char = merged_chars[-1] if merged_chars else None
                if last_char and last_char.is_alphanumeric:
                    last_char.value += char.value
                    continue
            merged_chars.append(char)

        return "".join(
            [
                (
                    MI(MIMathMlElement(char.value)).convert()
                    if not char.is_alphanumeric
                    else TextCommand(attributes.get("mathvariant")).apply(char.value)
                )
                for char in merged_chars
            ]
        )


class TagTypes(Enum):
    Under = 1
    Over = 2


class UnderOverSetter:
    def __init__(self, ty: TagTypes):
        self._type = ty

    def apply(self, content: str, accent: str) -> str:
        latex_accents = ["\\hat", "\\bar", "\\underbrace", "\\overbrace"]
        return (
            f"{accent}{{{content}}}"
            if accent in latex_accents
            else f"{self._default_command}{{{accent}}}{{{content}}}"
        )

    @property
    def _default_command(self) -> str:
        return "\\underset" if self._type == TagTypes.Under else "\\overset"


class GenericUnderOver(ToLaTeXConverter):
    def __init__(self, math_element: MathMLElement, adapter):
        self._math_element = math_element
        self._adapter = adapter

    def convert(self) -> str:
        name = self._math_element.name
        children = self._math_element.children
        children_length = len(children)

        if children_length != 2:
            raise InvalidNumberOfChildrenError(name, 2, children_length)

        content = self._adapter.to_latex_converter(children[0]).convert()
        accent = self._adapter.to_latex_converter(children[1]).convert()

        return self._apply_command(content, accent)

    def _apply_command(self, content: str, accent: str) -> str:
        ty = (
            TagTypes.Under
            if re.search(r"under", self._math_element.name)
            else TagTypes.Over
        )
        return UnderOverSetter(ty).apply(content, accent)


class MUnderover(ToLaTeXConverter):
    def __init__(self, math_element: MathMLElement, adapter):
        self._math_element = math_element
        self._adapter = adapter

    def convert(self) -> str:
        name = self._math_element.name
        children = self._math_element.children
        children_length = len(children)

        if children_length != 3:
            raise InvalidNumberOfChildrenError(name, 3, children_length)

        base = self._adapter.to_latex_converter(children[0]).convert()
        under_content = self._adapter.to_latex_converter(children[1]).convert()
        over_content = self._adapter.to_latex_converter(children[2]).convert()

        return f"{base}_{{{under_content}}}^{{{over_content}}}"


class MMultiscripts(ToLaTeXConverter):
    def __init__(self, math_element: MathMLElement, adapter):
        self._math_element = math_element
        self._adapter = adapter
        self._parenthesis_wrapper = ParenthesisWrapper()

    def convert(self) -> str:
        name = self._math_element.name
        children = self._math_element.children
        children_length = len(children)

        if children_length < 3:
            raise InvalidNumberOfChildrenError(name, 3, children_length, "at least")

        base_content = self._adapter.to_latex_converter(children[0]).convert()

        return (
            self._prescript_latex()
            + self._wrap_in_parenthesis_if_there_is_space(base_content)
            + self._postscript_latex()
        )

    def _prescript_latex(self) -> str:
        children = self._math_element.children

        if len(children) >= 4 and self._is_prescripts(children[1]):
            sub = children[2]
            sup = children[3]
        elif len(children) >= 6 and self._is_prescripts(children[3]):
            sub = children[4]
            sup = children[5]
        else:
            return ""

        sub_latex = self._adapter.to_latex_converter(sub).convert()
        sup_latex = self._adapter.to_latex_converter(sup).convert()

        return f"\\_{{{sub_latex}}}^{{{sup_latex}}}"

    def _postscript_latex(self) -> str:
        children = self._math_element.children

        if len(children) < 3 or (
            len(children) >= 1 and self._is_prescripts(children[1])
        ):
            return ""

        sub = children[1]
        sup = children[2]

        sub_latex = self._adapter.to_latex_converter(sub).convert()
        sup_latex = self._adapter.to_latex_converter(sup).convert()

        return f"_{{{sub_latex}}}^{{{sup_latex}}}"

    def _wrap_in_parenthesis_if_there_is_space(self, val: str) -> str:
        if not re.search(r"\s+", val):
            return val
        return self._parenthesis_wrapper.wrap(val)

    def _is_prescripts(self, child: MathMLElement) -> bool:
        return child.name == "mprescripts"



from typing import List, Dict
from xml.dom.minidom import Element as DOMElement, Node

from ...el_to_tex.protocols import MathMLElement, GenericMathMLElement


class ElementsToMathMLAdapter:
    def convert(self, els: List[DOMElement]) -> List[MathMLElement]:
        """
        Converts a list of DOM Elements to a list of MathMLElement instances.

        Args:
            els (List[DOMElement]): A list of DOM Elements to convert.

        Returns:
            List[MathMLElement]: A list of converted MathMLElement instances.
        """
        return [self._convert_element(el) for el in els if el.tagName is not None]

    def _convert_element(self, el: DOMElement) -> MathMLElement:
        attributes = self._convert_element_attributes(el.attributes)
        has_children = self._has_element_child(el)
        value = (
            "" if has_children else (el.firstChild.nodeValue if el.firstChild else "")
        )

        children = (
            self.convert(
                [
                    child
                    for child in el.childNodes
                    if child.nodeType == Node.ELEMENT_NODE
                ]
            )
            if has_children
            else []
        )

        return GenericMathMLElement(
            name=el.tagName,
            value=value.strip(),
            children=children,
            attributes=attributes,
        )

    def _convert_element_attributes(self, attributes) -> Dict[str, str]:
        attr_dict = {}
        for attr in attributes.values():
            # If the attribute value is the same as the attribute name, set it to an empty string
            attr_value = "" if attr.value == attr.name else attr.value
            attr_dict[attr.name] = attr_value
        return attr_dict

    def _has_element_child(self, el: DOMElement) -> bool:
        return any(child.nodeType == Node.ELEMENT_NODE for child in el.childNodes)

    def _is_there_any_no_text_node(self, children: List[Node]) -> bool:
        return any(child.nodeName != "#text" for child in children)



from xml.parsers.expat import ExpatError
import re


class ErrorHandler:
    def __init__(self):
        self.error_locator = {}

    def fix_error(self, xml: str, error: Exception) -> str:
        if not isinstance(error, ExpatError):
            return xml

        return self._fix_missing_attribute(error, xml)

    def _fix_missing_attribute(self, error: ExpatError, xml: str) -> str:
        result = self._fix_missing_attribute_value(error, xml)
        if result != xml:
            return result

        pattern = self._math_generic_missing_value()
        counter = 0
        while re.search(pattern, xml) and counter < 5:
            counter += 1
            # Replace the matched pattern by removing the missing attribute value
            result = re.sub(pattern, r"\1\3", xml)
        return result

    def _fix_missing_attribute_value(self, error: ExpatError, xml: str):
        error_code = error.code
        if error_code != 4:
            return xml

        error_line = error.lineno
        error_offset = error.offset

        goal_line = xml.split("\n")[error_line - 1]
        word = self._get_word_at_str_pos(goal_line, error_offset)

        pattern = self._match_missing_value_for_attribute(word)
        return re.sub(pattern, "", xml)

    def _match_missing_value_for_attribute(self, attribute: str) -> re.Pattern:
        """
        Constructs a regex pattern to find the specified attribute
        without a value (i.e., attribute= not followed by " or ').
        """
        escaped_attribute = re.escape(attribute)
        regex_pattern = (
            rf'({escaped_attribute}=(?!(["\'])))|({escaped_attribute}(?!(["\'])))'
        )
        return re.compile(regex_pattern, re.MULTILINE)

    def _math_generic_missing_value(self) -> re.Pattern:
        """
        Constructs a generic regex pattern to find any attribute
        without a value within a math element.
        """
        # This pattern matches:
        # Group 1: <... (anything up to a space)
        # Group 2: attribute= not followed by " or '
        # Group 3: ...> (anything up to >)
        return re.compile(r'(<.* )(\w+=(?!(["\'])))?(.*>)', re.MULTILINE)

    def _get_word_at_str_pos(self, string: str, pos: int) -> str:
        """
        Returns the word at the specified position in the string.
        """
        start = pos
        end = pos
        while start > 0 and string[start - 1] != " ":
            start -= 1
        while end < len(string) and string[end] != " " and string[end] != ">":
            end += 1
        return string[start:end]



from abc import ABC, abstractmethod


class MathMLElement(ABC):
    @property
    @abstractmethod
    def name(self) -> str:
        pass

    @property
    @abstractmethod
    def value(self) -> str:
        pass

    @property
    @abstractmethod
    def children(self) -> list:
        pass

    @property
    @abstractmethod
    def attributes(self) -> dict:
        pass


class ToLaTeXConverter(ABC):
    @abstractmethod
    def convert(self) -> str:
        pass


class GenericMathMLElement(MathMLElement):
    def __init__(self, name: str, value: str, children: list, attributes: dict):
        self._name = name
        self._value = value
        self._children = children
        self._attributes = attributes

    @property
    def name(self) -> str:
        return self._name

    @property
    def value(self) -> str:
        return self._value

    @property
    def children(self) -> list:
        return self._children

    @property
    def attributes(self) -> dict:
        return self._attributes


class VoidMathMLElement(MathMLElement):
    @property
    def name(self) -> str:
        return "void"

    @property
    def value(self) -> str:
        return ""

    @property
    def children(self) -> list:
        return []

    @property
    def attributes(self) -> dict:
        return {}


class MIMathMlElement(MathMLElement):
    def __init__(self, value: str):
        self._value = value

    @property
    def name(self) -> str:
        return "mi"

    @property
    def value(self) -> str:
        return self._value

    @property
    def children(self) -> list:
        return []

    @property
    def attributes(self) -> dict:
        return {}


class InvalidNumberOfChildrenError(Exception):
    def __init__(
        self,
        tag_name: str,
        expected_number_of_child: int,
        current_number_of_child: int,
        comparison: str = "exactly",
    ):
        super().__init__(
            f"{tag_name} tag must have {comparison} {expected_number_of_child} children. It's actually {current_number_of_child}"
        )
        self.name = "InvalidNumberOfChildrenError"



from .adapter import XmlToMathMLAdapter
from .services.error_handler import ErrorHandler
from .services.xml_elements_to_mathml_element import (
    ElementsToMathMLAdapter,
)


def make_xml_to_mathml_adapter() -> XmlToMathMLAdapter:
    xml_elements_to_mathml_element_adapter = ElementsToMathMLAdapter()
    error_handler = ErrorHandler()

    return XmlToMathMLAdapter(xml_elements_to_mathml_element_adapter, error_handler)



from .protocols import MathMLElement, VoidMathMLElement
from .usecases import (
    Void,
    MI,
    MO,
    MN,
    Math,
    MSup,
    GenericSpacingWrapper,
    MSqrt,
    MFenced,
    MTable,
    MTr,
    MSub,
    MFrac,
    MRoot,
    MAction,
    MEnclose,
    MError,
    MSubsup,
    MText,
    GenericUnderOver,
    MUnderover,
    MMultiscripts,
)


class MathMLElementToLatexConverterAdapter:
    def to_latex_converter(self, mathml_element: MathMLElement):
        name = mathml_element.name or VoidMathMLElement().name
        converter = self._from_mathml_element_to_latex_converter().get(
            name, self._make_generic_spacing_wrapper
        )

        return converter(mathml_element)

    def _from_mathml_element_to_latex_converter(self):
        return {
            "math": self._make_math_tag_converter,
            "mi": MI,
            "mo": MO,
            "mn": MN,
            "msqrt": self._make_msqrt_tag_converter,
            "mfenced": self._make_mfenced_tag_converter,
            "mfrac": self._make_mfrac_tag_converter,
            "mroot": self._make_mroot_tag_converter,
            "maction": self._make_maction_tag_converter,
            "menclose": self._make_menclose_tag_converter,
            "merror": self._make_merror_tag_converter,
            "mphantom": Void,
            "msup": self._make_msup_tag_converter,
            "msub": self._make_msub_tag_converter,
            "msubsup": self._make_msubsup_tag_converter,
            "mmultiscripts": self._make_mmultiscripts_tag_converter,
            "mtext": MText,
            "munderover": self._make_munderover_tag_converter,
            "mtable": self._make_mtable_tag_converter,
            "mtr": self._make_mtr_tag_converter,
            "mover": self._make_generic_under_over_converter,
            "munder": self._make_generic_under_over_converter,
            "mrow": self._make_generic_spacing_wrapper,
            "mpadded": self._make_generic_spacing_wrapper,
        }

    def _make_math_tag_converter(self, math_element: MathMLElement) -> Math:
        return Math(math_element, self)

    def _make_msup_tag_converter(self, math_element: MathMLElement) -> MSup:
        return MSup(math_element, self)

    def _make_msub_tag_converter(self, math_element: MathMLElement) -> MSub:
        return MSub(math_element, self)

    def _make_msubsup_tag_converter(self, math_element: MathMLElement) -> MSubsup:
        return MSubsup(math_element, self)

    def _make_msqrt_tag_converter(self, math_element: MathMLElement) -> MSqrt:
        return MSqrt(math_element, self)

    def _make_mfenced_tag_converter(self, math_element: MathMLElement) -> MFenced:
        return MFenced(math_element, self)

    def _make_mtable_tag_converter(self, math_element: MathMLElement) -> MTable:
        return MTable(math_element, self)

    def _make_mtr_tag_converter(self, math_element: MathMLElement) -> MTr:
        return MTr(math_element, self)

    def _make_generic_under_over_converter(
        self, math_element: MathMLElement
    ) -> GenericUnderOver:
        return GenericUnderOver(math_element, self)

    def _make_munderover_tag_converter(self, math_element: MathMLElement) -> MUnderover:
        return MUnderover(math_element, self)

    def _make_mfrac_tag_converter(self, math_element: MathMLElement) -> MFrac:
        return MFrac(math_element, self)

    def _make_mroot_tag_converter(self, math_element: MathMLElement) -> MRoot:
        return MRoot(math_element, self)

    def _make_maction_tag_converter(self, math_element: MathMLElement) -> MAction:
        return MAction(math_element, self)

    def _make_menclose_tag_converter(self, math_element: MathMLElement) -> MEnclose:
        return MEnclose(math_element, self)

    def _make_merror_tag_converter(self, math_element: MathMLElement) -> MError:
        return MError(math_element, self)

    def _make_mmultiscripts_tag_converter(
        self, math_element: MathMLElement
    ) -> MMultiscripts:
        return MMultiscripts(math_element, self)

    def _make_generic_spacing_wrapper(
        self, math_element: MathMLElement
    ) -> GenericSpacingWrapper:
        return GenericSpacingWrapper(math_element, self)



from typing import List

class Wrapper:
    def __init__(self, op: str, close: str):
        self._open = op
        self._close = close

    def wrap(self, val: str) -> str:
        return self._open + val + self._close

class ParenthesisWrapper(Wrapper):
    def __init__(self):
        super().__init__('\\left(', '\\right)')

    def wrap(self, val: str) -> str:
        return super().wrap(val)

    def wrap_if_more_than_one_char(self, val: str) -> str:
        if len(val) <= 1:
            return val
        return self.wrap(val)

class BracketWrapper(Wrapper):
    def __init__(self):
        super().__init__('{', '}')

    def wrap(self, val: str) -> str:
        return super().wrap(val)

class GenericWrapper(Wrapper):
    def __init__(self, op: str, close: str):
        open_cmd = f'\\left{op}'
        close_cmd = f'\\right{close}'
        super().__init__(open_cmd, close_cmd)

    def wrap(self, val: str) -> str:
        return super().wrap(val)

class JoinWithManySeparators:
    def __init__(self, separators: list):
        self._separators = separators

    def join(self, arr: List[str]) -> str:
        acc = ''
        for index, current_str in enumerate(arr):
            separator = '' if index == len(arr) - 1 else self._get(index)
            acc += current_str + separator
        return acc


    def _get(self, index: int) -> str:
        if index < len(self._separators):
            return self._separators[index]

        return self._separators[-1] if len(self._separators) > 0 else ','






from typing import List
from xml.dom.minidom import parseString, Element
import re

from ..el_to_tex.protocols import MathMLElement
from .services.error_handler import ErrorHandler
from .services.xml_elements_to_mathml_element import (
    ElementsToMathMLAdapter,
)


class XmlToMathMLAdapter:
    def __init__(
        self, elements_converter: ElementsToMathMLAdapter, error_handler: ErrorHandler
    ):
        self._xml = ""
        self._elements_converter = elements_converter
        self._error_handler = error_handler

    def convert(self, xml: str) -> List[MathMLElement]:
        self._xml = self._remove_line_breaks(xml)
        self._xml = self._remove_ms_word_prefixes(self._xml)

        return self._elements_converter.convert(self._mathml_elements)

    def _fix_error(self, error: Exception) -> None:
        self._xml = self._error_handler.fix_error(self._xml, error)

    def _remove_line_breaks(self, xml: str) -> str:
        line_break_pattern = re.compile(r"\n|\r\n|\r")
        return line_break_pattern.sub("", xml)

    def _remove_ms_word_prefixes(self, xml: str) -> str:
        ms_word_prefix_pattern = re.compile(r"mml:")
        return ms_word_prefix_pattern.sub("", xml)

    @property
    def _mathml_elements(self) -> List[Element]:
        exception = None

        for _ in range(5):
            try:
                dom = parseString(self._xml)
                elements = dom.getElementsByTagName("math")
                return list(elements)
            except Exception as e:
                self._fix_error(e)
                exception = e

        if exception:
            raise exception
